"""
 Bootlace Game

"""
from __future__ import division
import pygame
import data
import random
import colorsys
import pickle
import time

try:
    import psyco
    psyco.full()
except ImportError:
    pass

# Utilities
def tile_surface(surf,img,y_offset=0.5,x_offsets=(0,),y_gap=0,x_gap=0):
    w,h = surf.get_size()
    tw,th = img.get_size()
    y = int(-y_offset*th)
    xoffs = [int(dx * tw) for dx in x_offsets]
    which_dx = 0
    while y < h:
        x = -xoffs[which_dx]
        while x < w:
            surf.blit(img,(x,y))
            x += tw + x_gap
        which_dx = (which_dx + 1) % len(xoffs)
        y += th + y_gap

if pygame.version.vernum >= (1,8,0):
    def coloured_copy(img,rgba,blend=pygame.BLEND_MULT):
        coloured = img.copy()
        coloured.fill(rgba,img.get_rect(),blend)
        return coloured
else:
    def coloured_copy(img,rgba,blend=None):
        return img.copy()
    
def random_hue():
    r,g,b = (int(255*i) for i in colorsys.hsv_to_rgb(random.random(),1,1))
    return r,g,b,255

# Game objects
class Timer(object):
    def __init__(self,rect,seconds):
        self.rect = pygame.Rect(rect)
        self.max_ms = self.time_left_ms = seconds*1000
        self.adjust()
    def replenish(self,amount=1.0):
        self.time_left_ms = min(self.time_left_ms + int(amount*self.max_ms),self.max_ms)
        self.adjust()
    def deplete(self,ms):
        if not ms: return
        self.time_left_ms = max(0,self.time_left_ms-ms)
        self.adjust()
    def adjust(self):
        x,y,w,h = self.rect
        t_ms = self.time_left_ms
        m_ms = self.max_ms
        frac = t_ms / m_ms
        bar_w = int((w-2) * frac)
        self.bar = pygame.Rect(x+1,y+1,bar_w,h-2)
        self.colour = (128, int(255*frac),int(128*frac))
    def draw(self,screen):
        screen.fill((32,32,32),self.rect)
        screen.fill(self.colour,self.bar)
        return [self.rect]
    def __nonzero__(self):
        return self.time_left_ms

class Score(object):
    def __init__(self,rect,max_score,hiscore):
        self.score = 0
        self.hiscore = hiscore
        if hiscore >= max_score:
            max_score = (hiscore * 3) // 2
        self.max_score = max_score
        self.font = pygame.font.Font(None,18)
        nw,nh = self.font.size("99999")
        x,y,w,h = rect
        self.rect = pygame.Rect(x+nw+5, y, w-nw-5,h)
        self.numpos = x, y-((nh-h)//2)
        self.adjust()
    def increment(self,amount):
        self.score += amount
        self.adjust()
    def adjust(self):
        x,y,w,h = self.rect
        s = self.score
        hi = self.hiscore
        m = self.max_score
        frac = s / m
        hifrac = hi / m
        if s > hi:
            self.colour = (255,248,140)
        else:
            self.colour = (60,80,200)
        bar_w = int((w-2) * frac)
        hibar_w = int((w-2) * hifrac)
        self.bar = pygame.Rect(x+1,y+1,bar_w,h-2)
        self.hibar = pygame.Rect(x+1,y+2,hibar_w,h-4)
        self.text = self.font.render("%05d" % s, True, self.colour,(32,32,32))
    def draw(self,screen):
        screen.fill((32,32,32),self.rect)
        if (self.score < self.hiscore):
            screen.fill((255,0,0),self.hibar)
        screen.fill(self.colour, self.bar)
        trect = screen.blit(self.text,self.numpos)
        return [self.rect, trect]

# Directions
North = (0,-1)
East = (1,0)
West = (-1,0)
South = (0,1)

def opposite(direction):
    dx,dy = direction
    return -dx, -dy

def direction(coordsa,coordsb):
    xa,ya = coordsa
    xb,yb = coordsb
    return xb-xa,yb-ya

class StringPiece(object):
    ends = ()
    name = None
    def _get_filename(self):
        return "string_" + self.name.lower() + ".png"
    def _get_cursor_filename(self):
        return "string_" + self.name.lower() + ".xbm"

    filename = property(_get_filename)
    cur_filename = property(_get_cursor_filename)

    def can_stack(self,other_pieces):
        return not other_pieces

    def set_cursor(self):
        cur = data.load_cursor(self.cur_filename)
        pygame.mouse.set_cursor(*cur)

class Eye(object):
    name = "Eye"
    filename = "eye.png"
    
class StringV(StringPiece):
    ends = (North,South)
    name = "V"
    def can_stack(self,other_pieces):
        return (not other_pieces or
                (len(other_pieces) == 1 and
                 other_pieces[0].name == "H"))

class StringH(StringPiece):
    ends = (East,West)
    name = "H"
    def can_stack(self,other_pieces):
        return (not other_pieces or
                (len(other_pieces) == 1 and
                 other_pieces[0].name == "V"))

class StringNW(StringPiece):
    ends = (North, West)
    name = "NW"

class StringNE(StringPiece):
    ends = (North, East)
    name = "NE"

class StringSW(StringPiece):
    ends = (South,West)
    name = "SW"

class StringSE(StringPiece):
    ends = (South, East)
    name = "SE"

class Blank(object):
    filename = "back_grid.png"

all_string_pieces = [
    StringH, StringV, StringH, StringV,
    StringNW, StringNE, StringSW, StringSE
    ]

# Pieces only added for decorative effect
class EndN(StringPiece):
    ends = (North,)
    name = "N"
class EndW(StringPiece):
    ends = (West,)
    name = "W"
class EndE(StringPiece):
    ends = (East,)
    name = "E"
class EndS(StringPiece):
    ends = (South,)
    name = "S"

end_piece_from_direction = {
    North:EndS,
    South:EndN,
    East:EndW,
    West:EndE
    }

class Tile(object):
    def __init__(self,piece,col,row,colour=None):
        img = data.load_image(piece.filename)
        if colour:
            img = coloured_copy(img,colour)
        self.img = img
        self.piece = piece
        tw = self.img.get_width()
        self.row, self.col = row,col
        self.rect = self.img.get_rect().move(tw*col,tw*row)
    def draw(self,surface):
        surface.blit(self.img,self.rect)
        return self.rect

class GameGrid(object):
    def __init__(self,screen):
        self.screen = screen
        self.grid_img = data.load_image("back_grid.png")
        self.tile_width = tw = self.grid_img.get_width()
        self.colourise = random_hue()
        w = 8 * tw
        h = 6 * tw
        x = (screen.get_width() - w) // 2
        y = (screen.get_height() - h) // 2
        self.grid_surface = pygame.Surface((w,h))
        self.rect = pygame.Rect(x,y,w,h)
        tile_surface(self.grid_surface,self.grid_img,y_offset=0)
        self.tiles_placed = {} # (col,row):[Tile]
        self.eyes = set() # set of (col,row)
        self.threaded_tiles = set() # set of (col,row)
        self.allow_scissors = False

    def get(self,col,row):
        if 0 <= col <= 8 and 0 <= row <= 6:
            return self.tiles_placed.setdefault((col,row),list())
        return None

    def empties(self):
        return [(c,r) for r in xrange(6) for c in xrange(8)
                if ((c,r) not in self.eyes and
                    (c,r) not in self.tiles_placed)]

    def place_string(self,tile):
        score = 0
        col,row = tile.col, tile.row
        ts = self.get(col,row)
        ts.append(tile)
        tile.draw(self.grid_surface)
        end1,end2 = tile.piece.ends
        string1 = self.follow(col,row,end1)
        string2 = self.follow(col,row,end2)
        if string1 and string2:
            string1.reverse()
            all_cells = string1 +[(col,row)] + string2
            self.thread_into_eyes(all_cells)
            self.allow_scissors = False
            score = 20 * len(all_cells)
        else:
            score = len(string1) + len(string2)
            self.allow_scissors = True
        return score

    def follow(self,col,row,end):
        dx,dy = end
        newx,newy = col+dx, row+dy
        if (newx,newy) in self.eyes:
            return [(newx,newy)]
        ts = self.get(newx,newy)
        if ts is None: return []
        joining_end = opposite(end)
        for t in ts:
            t_ends = list(t.piece.ends)
            try:
                t_ends.remove(joining_end)
                other_end = t_ends[0]
                rest = self.follow(newx,newy,other_end)
                if rest:
                    return [(newx,newy)] + rest
                continue
            except (ValueError, IndexError):
                continue
        return []

    def thread_into_eyes(self,coords):
        col,row = coords[0]
        endpiece = end_piece_from_direction[direction(coords[1],(col,row))]()
        t= Tile(endpiece,col,row,self.colourise)
        t.draw(self.grid_surface)
        ts = self.get(col,row)
        ts.append(t)
        col,row = coords[-1]
        endpiece = end_piece_from_direction[direction(coords[-2],(col,row))]()
        t = Tile(endpiece,col,row,self.colourise)
        t.draw(self.grid_surface)
        ts = self.get(col,row)
        ts.append(t)
        self.threaded_tiles |= set(coords)
        unthreaded = set((c,r) for c in xrange(8) for r in xrange(6)) - self.threaded_tiles
        for col,row in unthreaded:
            if (col,row) in self.tiles_placed:
                del self.tiles_placed[col,row]
            blank = Tile(Blank(),col,row) #probably want a cleaner way to do this!
            blank.draw(self.grid_surface)
        self.place_eyes(1)
        self.colourise = random_hue()

    def place_eyes(self,n=2):
        places = self.empties()
        while n > 0 and places:
            c,r = random.choice(places)
            n -= 1
            self.place_eye(c,r)
            places.remove((c,r))

    def place_eye(self,col,row):
        self.eyes.add((col,row))
        eye = Tile(Eye(),col,row)
        eye.draw(self.grid_surface)

    def draw(self):
        scr = self.screen
        scr.blit(self.grid_surface, self.rect)
        return [self.rect]

    def can_place(self, piece, col, row):
        if (col,row) in self.eyes:
            return False
        ts = self.get(col,row)
        if ts is None:
            return False
        pieces = [t.piece for t in ts]
        return piece.can_stack(pieces)

    def can_snip(self,col,row):
        if (col,row) in (self.eyes | self.threaded_tiles):
            return False
        ts = self.get(col,row)
        if not ts: # None or empty
            return False
        return True

    def snip(self,col,row):
        ts = self.get(col,row)
        ts[-1:] = [] # chop off top one
        blank = Tile(Blank(),col,row)
        blank.draw(self.grid_surface)
        for t in ts:
            t.draw(self.grid_surface)
            

    def grid_pos(self,mouse_pos):
        mx,my = mouse_pos
        x,y,w,h = self.rect
        tw = self.tile_width
        col = (mx - x) // tw
        row = (my - y) // tw
        if 0 <= col <= 8 and 0 <= row <= 6:
            return col,row
        return None

class GameScreen(object):
    def __init__(self,screen):
        self.screen = screen
        self.timer = Timer((100,5,600,10),60)
        self.grid = GameGrid(screen)
        try:
            f = data.load("hiscore")
            hiscore = pickle.load(f)
        except IOError:
            hiscore = 100
        self.score = Score((100,585,600,12),1000,hiscore)
        self.game_over = False
        self.back_img = data.load_image("back_sakura.png")
        self.grid.place_eyes()
        self.next_piece = random.choice(all_string_pieces)()
        self.scissors = False
    def animate(self,ms):
        self.timer.deplete(ms)
        if not self.timer: self.game_over = True
    def draw_first(self):
        tile_surface(self.screen, self.back_img,x_offsets=(0.33,))
        self.timer.draw(self.screen)
        pygame.display.flip()
    def draw(self):
        screen = self.screen
        rects = []
        rects.extend(self.timer.draw(screen))
        rects.extend(self.score.draw(screen))
        rects.extend(self.grid.draw())
        pygame.display.update(rects)
    def events(self,events):
        for e in events:
            if e.type == pygame.QUIT:
                self.game_over = True
            elif e.type == pygame.KEYDOWN and e.key == pygame.K_ESCAPE:
                self.game_over = True
            elif e.type == pygame.MOUSEBUTTONDOWN:
                self.click(e.pos, e.button)
            elif e.type == pygame.MOUSEMOTION:
                self.motion(e.pos)
            else:
                pass
    def click(self,pos,button):
        if button == 3 and self.scissors:
            self.scissors = False
            self.next_piece = random.choice(all_string_pieces)()
            return
        elif button != 1:
            return
        grid_pos = self.grid.grid_pos(pos)
        if not grid_pos:
            return
        if self.scissors:
            can_snip = self.grid.can_snip(*grid_pos)
            if not can_snip:
                return
            self.grid.snip(*grid_pos)
        else:
            piece = self.next_piece
            can_place = self.grid.can_place(piece,*grid_pos)
            if not can_place:
                return
            col,row = grid_pos
            tile = Tile(piece, col, row, self.grid.colourise)
            score_plus = self.grid.place_string(tile)
            self.score.increment(score_plus)
            if score_plus > 20:
                self.timer.replenish(0.1)
            elif score_plus:
                self.timer.replenish(0.01)
        if random.random() < 0.1 and self.grid.allow_scissors:
            self.scissors = True
        else:
            self.scissors = False
            self.next_piece = random.choice(all_string_pieces)()
        
    def motion(self,pos):
        grid_pos = self.grid.grid_pos(pos)
        cursor_set = False
        if grid_pos:
            if self.scissors:
                if self.grid.can_snip(*grid_pos):
                    cur = data.load_cursor("scissors.xbm")
                    pygame.mouse.set_cursor(*cur)
                    cursor_set = True
            elif self.grid.can_place(self.next_piece,*grid_pos):
                self.next_piece.set_cursor()
                cursor_set = True
        if not cursor_set:
            pygame.mouse.set_cursor(*pygame.cursors.broken_x)

def run_game(screen):
    clock = pygame.time.Clock()
    gscr = GameScreen(screen)
    gscr.draw_first()
    while not gscr.game_over:
        ms = clock.tick(60)
        gscr.events(pygame.event.get())
        gscr.animate(ms)
        gscr.draw()
    pygame.mouse.set_cursor(*pygame.cursors.arrow)
    score, hiscore = gscr.score.score, gscr.score.hiscore
    run_hiscore(score,hiscore)
        
def run_hiscore(score, hiscore):
    font = pygame.font.Font(None,64)
    if score > hiscore:
        f = data.load("hiscore","wb")
        pickle.dump(score,f)
        nhs = font.render("New High Score : %d" % score,
                          True,(255,248,140),(0,0,0))
    else:
        nhs = font.render("Score %d" % score,
                          True,(255,248,140),(0,0,0))
    w,h = nhs.get_size()
    pygame.display.get_surface().blit(nhs,((800-w)//2, (600-h)//2))
    pygame.display.flip()
    time.sleep(2) # ignore overzealous frantic clicking
    pygame.event.clear()
    pygame.event.wait()
    
def run_title(screen):
    pygame.display.set_caption("bootlace")
    back = data.load_image("back_sakura.png")
    tile_surface(screen,back,x_offsets=(0,0.5))
    title = data.load_image("title.png")
    instructions = data.load_image("instructions.png")
    w = screen.get_width()
    x = (w - title.get_width()) // 2
    screen.blit(title,(x,100))
    x = (w - instructions.get_width()) // 2
    screen.blit(instructions,(x,400))
    pygame.display.flip()

    clock = pygame.time.Clock()
    key_responses = {
        pygame.K_RETURN: True,
        pygame.K_ESCAPE: False,
        }
    while True:
        dummy = clock.tick(60)
        # animate?
        e = pygame.event.poll()
        if e.type == pygame.QUIT:
            return False
        if e.type == pygame.KEYDOWN:
            try:
                return key_responses[e.key]
            except:
                pass
        elif e.type == pygame.MOUSEBUTTONDOWN:
            return True
